/** \file almlib.c
 * \brief TINE Local Alarm Server plus API routines
 *
 * The API routines offered here pertain directly to the TINE Local Alarm Server maintained
 * on the running server.  A server can of course run without setting alarms.  However a 
 * control system will suffer if important information relating to the health of the system
 * is surpressed.  One can set a certain category of alarms (threshold alarms) without issuing
 * any API call, but instead supplying a startup configuration file 'almwatch.csv' with 
 * instructions as to which properties to call and into what range the data returned must fall
 * in order to function properly.  Otherwise calls to ClearAlarm() and SetAlarm() for instance
 * should be made inside a server's IO loop where necessary.
 *
 */
#ifndef NO_DOXYGEN
#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include "srvcore.h"
#include "errors.h"
#include "dbtools.h"
#include "almlib.h"
#include "tinetype.h"

#define findDeviceNr(eqm,dev) GetDeviceNumber(eqm,dev)
extern char FecDBpath[80];
short ftoi(char *fmt);

int numStoredAlarms = 0;
int NSeverityCodes = 0;
int AlarmServerInitialized = 0;
int AlarmTestState = 0;
int ALMdebug = 0;
AWSList *almWatchList = NULL;
int nAlmWatchList = 0;
int gRequireReadFromCAS = FALSE;
int gIsCentralAlarmServer = FALSE;
int populateADS(char *fecname,char *eqmname,DBLNKLIST *fecList,ADSIS **adsis);
int populateAWS(char *fecname,DBLNKLIST *fecList,AWSIS **awsis);

TINE_EXPORT void isCentralAlarmServer(int value) { gIsCentralAlarmServer = value; }
static int alarmInfoTableFile = FALSE;
int gAlarmSurveyInterval = DEFAULT_ALARM_SURVEY_INTERVAL;
TINE_EXPORT int SetAlarmSurveyInterval(UINT32 interval)
{
  if (interval < gSystemTick) return out_of_range;
  gAlarmSurveyInterval = interval;
  return 0;
}
TINE_EXPORT int GetAlarmSurveyInterval(void)
{
  return gAlarmSurveyInterval;
}

#define NEW_ALARM_WINDOW 3

#define idx_too_high      0
#define idx_warn_too_high 1
#define idx_too_low       2
#define idx_warn_too_low  3
#define idx_invalid_data  4

ADS almWatchTableADS[] = 
{
  { "Value Too HIGH",value_too_high,0,0,14,BFMT(CF_FLOAT),1,0,0,"measured value is above threshold","","last measured value","" },
  { "Value Too High",warn_too_high,0,0,12,BFMT(CF_FLOAT),1,0,0,"measured value is above warn threshold","","last measured value",""},
  { "Value Too LOW",value_too_low,0,0,14,BFMT(CF_FLOAT),1,0,0,"measured value is below threshold","","last measured value",""},
  { "Value Too Low",warn_too_low,0,0,12,BFMT(CF_FLOAT),1,0,0,"measured value is below warn threshold","","last measured value",""},
  { "Invalid Data",invalid_data,0,0,12,BFMT(CF_LONG),1,0,0,"measured value not at set point","","last measured value",""}
};
int csvAlmWatchFileGetFormat(char *fmtstr);
CSVCOLUMN csvAlmWatchFileColumns[] = 
{
  { "LOCAL_NAME","",0,CF_TEXT,EQM_NAME_SIZE,(BYTE *)OFFSETIN(AWSIS,eqm),NULL, 0 },
  { "DEVICE_NAME","",0,CF_TEXT,DEVICE_NAME_SIZE,(BYTE *)OFFSETIN(AWSIS,dev),NULL, 0 },
  { "PROPERTY","",0,CF_TEXT,PROPERTY_NAME_SIZE,(BYTE *)OFFSETIN(AWSIS,prp),NULL, 0 },
  { "SIZE","",0,CF_LONG,1,(BYTE *)OFFSETIN(AWSIS,siz),NULL, 0 },
  { "FORMAT","",0,CF_TEXT,64,(BYTE *)OFFSETIN(AWSIS,alarmDataFormatString),NULL, 0 },
  { "SEVERITY","",0,CF_LONG,1,(BYTE *)OFFSETIN(AWSIS,sev),NULL, 0 },
  { "SEVERITY_HIGH","",0,CF_LONG,1,(BYTE *)OFFSETIN(AWSIS,hisev),NULL, 0 },
  { "SEVERITY_LOW","",0,CF_LONG,1,(BYTE *)OFFSETIN(AWSIS,losev),NULL, 0 },
  { "SEVERITY_HIGHWARN","",0,CF_LONG,1,(BYTE *)OFFSETIN(AWSIS,hiwarnsev),NULL, 0 },
  { "SEVERITY_LOWWARN","",0,CF_LONG,1,(BYTE *)OFFSETIN(AWSIS,lowarnsev),NULL, 0 },
  { "ALARM_SYSTEM","0",0,CF_LONG,1,(BYTE *)OFFSETIN(AWSIS,asys),NULL, 0 },
  { "MASK","0",0,CF_LONG,1,(BYTE *)OFFSETIN(AWSIS,mask),NULL, 0 },
  { "NORMAL","0",0,CF_LONG,1,(BYTE *)OFFSETIN(AWSIS,normal),NULL, 0 },
  { "COUNT_THRESHOLD","0",0,CF_LONG,1,(BYTE *)OFFSETIN(AWSIS,cntThreshold),NULL, 0 },
  { "HIGH","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(AWSIS,hi),NULL, 0 },
  { "LOW","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(AWSIS,lo),NULL, 0 },
  { "HIGHWARN","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(AWSIS,hiwarn),NULL, 0 },
  { "LOWWARN","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(AWSIS,lowarn),NULL, 0 }
};
CSVDB csvAlmWatchFileDb =
{
  csvAlmWatchFileColumns,
  sizeof(csvAlmWatchFileColumns)/sizeof(csvAlmWatchFileColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(AWSIS)
};
int GetAlarmWatchTable(void)
{
  int i, cc = 0;
  char almfn[64];
  AWSIS *aws = NULL;
  AWSList *lst = NULL;

  if (isUsingXMLFecDatabase)
  {
    nAlmWatchList = populateAWS(gFecName,gFecInfoList,&aws);
    if (nAlmWatchList < 0) ccerr(-nAlmWatchList);
  }
  else
  {
    strncpy(almfn,ALMARM_WATCH_FILE,64);
    if (strlen(gFecDBSubDir) > 0) sprintf(almfn,"%s%s",gFecDBSubDir,ALMARM_WATCH_FILE);
    if ((cc=csvReadFile(FecDBpath,almfn,&csvAlmWatchFileDb,(void **)&aws)) != 0) ccerr(cc);
    nAlmWatchList = csvAlmWatchFileDb.siz;
  }
  for (i=0; i<nAlmWatchList; i++)
  {
    if ((lst=(AWSList *)SystemCalloc(1,sizeof(AWSList))) == NULL) ccerr(out_of_server_memory);
    aws[i].atyp = getArrayType(aws[i].alarmDataFormatString);
    aws[i].fmt = BFMT(ftoi(aws[i].alarmDataFormatString));
    if (aws[i].mask != 0) aws[i].chkNormal = TRUE;
    if (aws[i].hisev == 0) aws[i].hisev = aws[i].sev;
    if (aws[i].hiwarnsev == 0) aws[i].hiwarnsev = aws[i].hisev - 2;
    if (aws[i].losev == 0) aws[i].losev = aws[i].sev;
    if (aws[i].lowarnsev == 0) aws[i].lowarnsev = aws[i].losev - 2;
    memcpy(&lst->aws,&aws[i],sizeof(AWS));
    lst->nxt = almWatchList;
    almWatchList = lst;
  }

err:
  if (aws != NULL) SystemFree(aws);
  feclog("ALM: watch file : %s", cc ? erlst[cc] : "found");
  return cc;
}
static CLN almcln;
static CONSUMER almcon;
static ContractListStruct almcl;
void checkAlarmsInWatchTable(void)
{
  AWSList *aw;
  int j,cc,cnt;
  short devNr;
  ExportListStruct *el;
  char sDevName[DEVICE_NAME_SIZE+4];
  float fval;
  UINT32 ival;
  static BYTE almdata[ALARM_DATA_SIZE];
  static DTYPE din,dout;
  static int done = 0;

  if (!done) 
  { /* initialization */
    strncpy(almcln.userName,"LOCAL_ALARM",USERNAME_SIZE);
    almcon.client = &almcln;
    almcl.consumer = &almcon;
    almcl.nconsumers = 1;
    memset(&din,0,sizeof(DTYPE));
  }
  done = TRUE;
  for (aw=almWatchList; aw != NULL; aw = aw->nxt)
  {
    if ((el=getExportListItem(aw->aws.eqm)) == NULL) continue;
    if (el->eqmFcn == NULL) continue;
    if (el->eqmbusy) continue;
    if ((devNr=GetDeviceNumber(aw->aws.eqm,aw->aws.dev)) < 0)
    {
      if (tineDebug) dbglog("LAS: %s is not a registered device",aw->aws.dev);
      continue;
    }
    if (aw->aws.cntThreshold < 0) continue; /* disabled */
    dout.dArrayLength = aw->aws.siz;
    dout.dFormat = LFMT(aw->aws.fmt);
    dout.data.vptr = tmpWorkArea;
    dout.dTag[0] = 0;
    strncpy(sDevName,aw->aws.dev,DEVICE_NAME_SIZE);
    almcl.exp = el;
    almcl.compStatus = not_signalled;
    cc = callEqm(&almcl,sDevName,aw->aws.prp,&dout,&din,CA_READ|CA_ALARM);
    if (cc != 0)
    {
      if (tineDebug)
        dbglog("LAS: %s %s : <%d> %s",
               sDevName,aw->aws.prp,cc, cc > 0 && cc < numErr ? erlst[cc] : "undefined");
      continue;
    }
    cnt = 0;
    for (j=0; j<aw->aws.siz; j++)
    {
      if (aw->aws.atyp == AT_CHANNEL) devNr = j;
      if (aw->aws.atyp != AT_WAVEFORM && aw->aws.siz == el->EqmNumDevices) devNr = j; /* assume array of devices */
      switch (LFMT(aw->aws.fmt))
      {
        case CF_BYTE: 
          ival = (UINT32)dout.data.bptr[j]; 
          fval = (float)dout.data.bptr[j]; 
          break;
        case CF_SHORT:
          ival = (UINT32)dout.data.sptr[j]; 
          fval = (float)dout.data.sptr[j]; 
          break;
        case CF_LONG: 
          ival = (UINT32)dout.data.lptr[j]; 
          fval = (float)dout.data.lptr[j]; 
          break;
        case CF_FLOAT:
          ival = (UINT32)dout.data.fptr[j]; 
          fval = (float)dout.data.fptr[j]; 
          break;
        case CF_DOUBLE:
          ival = (UINT32)dout.data.dptr[j]; 
          fval = (float)dout.data.dptr[j]; 
          break;
        default: fval = 0;
      }
      if (aw->aws.chkNormal)
      {
        if (aw->aws.normal != (ival & aw->aws.mask))
        {
          cnt++;
          memcpy(almdata,&ival,4);
          aw->aws.cnt++;
          almWatchTableADS[idx_invalid_data].alarmSystem = aw->aws.asys;
          almWatchTableADS[idx_invalid_data].alarmSeverity = aw->aws.sev;
          if (aw->aws.cnt > aw->aws.cntThreshold) SetAlarm(aw->aws.eqm,devNr,invalid_data,almdata);
        }
      }
      else if (fval > aw->aws.hiwarn) 
      {
        cnt++;
        memcpy(almdata,&fval,4);
        aw->aws.cnt++;
        almWatchTableADS[idx_too_high].alarmSystem = aw->aws.asys;
        almWatchTableADS[idx_too_high].alarmSeverity = aw->aws.hisev;
        almWatchTableADS[idx_warn_too_high].alarmSystem = aw->aws.asys;
        almWatchTableADS[idx_warn_too_high].alarmSeverity = aw->aws.hiwarnsev;
        if (aw->aws.cnt > aw->aws.cntThreshold) 
          SetAlarm(aw->aws.eqm,devNr,fval > aw->aws.hi ? value_too_high : warn_too_high,almdata);
      }
      else if (fval < aw->aws.lowarn) 
      {
        cnt++;
        memcpy(almdata,&fval,4);
        aw->aws.cnt++;
        almWatchTableADS[idx_too_low].alarmSystem = aw->aws.asys;
        almWatchTableADS[idx_too_low].alarmSeverity = aw->aws.losev;
        almWatchTableADS[idx_warn_too_low].alarmSystem = aw->aws.asys;
        almWatchTableADS[idx_warn_too_low].alarmSeverity = aw->aws.lowarnsev;
        if (aw->aws.cnt > aw->aws.cntThreshold) 
          SetAlarm(aw->aws.eqm,devNr,fval < aw->aws.lo ? value_too_low : warn_too_low,almdata);
      }
      else 
      {
        if (aw->aws.cnt) /* if there is or was an alarm somewhere, run through RemoveAlarm() for the others */
        {
          RemoveAlarm(aw->aws.eqm,devNr,invalid_data);
          RemoveAlarm(aw->aws.eqm,devNr,value_too_high);
          RemoveAlarm(aw->aws.eqm,devNr,value_too_low);
          RemoveAlarm(aw->aws.eqm,devNr,warn_too_high);
          RemoveAlarm(aw->aws.eqm,devNr,warn_too_low);
        }
      }
    }
    /* only clear the watch table counter when a complete pass has no alarms */
    if (cnt == 0) aw->aws.cnt = 0;
  }
  return;
}
void getAlarmInfo(void)
{
  ExportListStruct *el;
  char str[64];

  for (el=ExportList; el != NULL; el=el->next)
  {
    if (el->adsTable != NULL) continue;
    if (getAlarmInfoTable(el->EqmName) < 0)
    {
      sprintf(str,"No alarm information for %s\n>",el->EqmName);
      feclog(str);
    }
  }
}
CSVCOLUMN csvAlarmInfoColumns[] = 
{
  { "ALARM_TAG","",0,CF_TEXT,ALARM_TAG_SIZE,(BYTE *)OFFSETIN(ADSIS,alarmTag),NULL,0 },
  { "ALARM_CODE","0",0,CF_LONG,1,(BYTE *)OFFSETIN(ADSIS,alarmCode),NULL,0 },
  { "ALARM_MASK","0",0,CF_LONG,1,(BYTE *)OFFSETIN(ADSIS,alarmMask),NULL,0 },
  { "SEVERITY","1",0,CF_SHORT,1,(BYTE *)OFFSETIN(ADSIS,alarmSeverity),NULL,0 },
  { "DATA_FORMAT","",0,CF_TEXT,64,(BYTE *)OFFSETIN(ADSIS,alarmSeverity),NULL,0 },
  { "DATA_ARRAYSIZE","0",0,CF_BYTE,1,(BYTE *)OFFSETIN(ADSIS,alarmDataFormatString),NULL,0 },
  { "ALARM_TEXT","not supplied",0,CF_TEXT,ALARM_TEXT_SIZE,(BYTE *)OFFSETIN(ADSIS,alarmText),NULL,0 },
  { "DEVICE_TEXT","not supplied",0,CF_TEXT,ALARM_TEXT_SIZE,(BYTE *)OFFSETIN(ADSIS,deviceText),NULL,0 },
  { "DATA_TEXT","not supplied",0,CF_TEXT,ALARM_TEXT_SIZE,(BYTE *)OFFSETIN(ADSIS,dataText),NULL,0 },
  { "URL","not supplied",0,CF_TEXT,ALARM_TEXT_LONGSIZE,(BYTE *)OFFSETIN(ADSIS,url),NULL,0 },
  { "ALARM_SYSTEM","0",0,CF_SHORT,1,(BYTE *)OFFSETIN(ADSIS,alarmSystem),NULL,0 }
};
CSVDB csvAlarmInfoDb =
{
  csvAlarmInfoColumns,
  sizeof(csvAlarmInfoColumns)/sizeof(csvAlarmInfoColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(ADSIS)
};
char *makeAlarmsFileName(char *eqm,int decorate)
{
  static char fname[64];
  char dname[32];

  if (decorate == 0)
  {
    sprintf(fname,"%salarms.csv",gFecDBSubDir);
  }
  else
  {
    memset(dname,0,sizeof(dname));
    strncpy(dname,eqm,EQM_NAME_SHORTSIZE);
    strlwr(dname);
#   if !defined(FS_SHORTNAMES)
    strcat(dname,"-alarms.csv");
#   else
    strcat(dname,"-a.csv");
#   endif
    sprintf(fname,"%s%s",gFecDBSubDir,dname);
  }
  return fname;
}
int getAlarmInfoTable(char *eqm)
{
  int i,n,cc=0,hasAlarmSystem=0,decorate=0;
  ADSIS *ads = NULL;
  ADSList *lst = NULL;
  char *fname=NULL;
  ExportListStruct *el;

  if ((el=getExportListItem(eqm)) == NULL) ccerr(non_existent_elem);

  if (isUsingXMLFecDatabase)
  {
    n = populateADS(gFecName,eqm,gFecInfoList,&ads);
    if (n < 0) ccerr(-n);
  }
  else
  {
retryEqmAlarmsFile:
    fname = makeAlarmsFileName(eqm,decorate);
    if ((cc=csvReadFile(FecDBpath,fname,&csvAlarmInfoDb,(void **)&ads)) == no_such_file && decorate++ == 0) 
      goto retryEqmAlarmsFile;
    if (cc != 0) ccerr(cc);
    n = csvAlarmInfoDb.siz;
  }
  for (i=0; i<n; i++) 
  {
    if ((lst=(ADSList *)SystemCalloc(1,sizeof(ADSList))) == NULL) ccerr(out_of_server_memory);
    ads[i].alarmDataFormat = BFMT(ftoi(ads[i].alarmDataFormatString));
    memcpy(&lst->ads,&ads[i],sizeof(ADS));
    lst->next = el->adsTable;
    el->adsTable = lst;
    if (ads[i].alarmSystem) hasAlarmSystem = 1;
  }
  el->alarmSystem = hasAlarmSystem;
  el->adsTableSize = n;
  feclog("read ALM file for %s",eqm);
  alarmInfoTableFile = TRUE;
err:
  if (cc && cc != no_such_file) feclog("getAlarmInfoTable : %s",erlst[cc]);
  if (ads != NULL) SystemFree(ads);
  return 0;
}
int isCASInUserList(char *eqm)
{
  ExportListStruct *el = getExportListItem(eqm);
  FecDataStruct *fec = NULL;
  NAME16 usrlst[16];
  int i, n = 16;
  static int casUnknown = FALSE;

  if (el == NULL) return FALSE;
  if (casUnknown) return FALSE;
  if ((fec=GetFecDataFromCache(el->EqmContext,"CAS")) == NULL) 
  { /* CAS not known so turn this check off */
    casUnknown = TRUE;
    return FALSE;
  }
  GetCallerInfo(STKEQM_NAME,usrlst,NULL,NULL,NULL,&n);
  for (i=0; i<16 && i<n; i++) if (!strnicmp(fec->Name,usrlst[i].name,16)) break;
  return i < 16 ? TRUE : FALSE;
}
/* TODO:  make this do something */
int RegisterSeverity(int almCode,int almSeverity)
{
  almCode = almCode;
  almSeverity = almSeverity;
  return 0;
}
/* TODO:  make this do something */
void enforceAlarm(short cc)
{
 /* method to make sure this alarm goes out */
  cc = cc;
  return;
}

void clearCASAlarmList(void)
{
  ExportListStruct *el;
  int cc = 0;
  char devName[256];

  for (el=ExportList; el != NULL; el=el->next)
  {
    if (strlen(el->EqmContext) == 0) strncpy(el->EqmContext,gDeviceContext,CONTEXT_NAME_SIZE);
    if (strlen(el->EqmSubSystem) == 0) strncpy(el->EqmSubSystem,gDeviceSubSystem,SUBSYSTEM_NAME_SIZE);
    if (!strcmp(el->EqmName,SRVEQM_NAME)) continue;
    sprintf(devName,"/%s/CAS/%s",el->EqmContext,el->EqmExportName);
    cc = ExecLink(devName,"REMOVEALARMS",NULL,NULL,CA_WRITE|CA_RETRY);
    if (cc == 0) gRequireReadFromCAS = TRUE; /* the CAS deems this FEC to be important */
  }
  feclog("ALM: clear CAS Alarm List : %s",erlst[cc]);
}
int initAlarmServer(void)
{
  feclog("ALM: initializing local alarm subsystem");
  GetAlarmWatchTable();
  if (!gIsCentralAlarmServer) clearCASAlarmList();
  AlarmServerInitialized = 1;
  return 0;
}
int verifyDeviceNr(ExportListStruct *el,short devNr)
{
 /* verify that the device number passed in the API is valid */
  if (el == NULL || devNr >= el->EqmNumDevices || devNr < 0) return -1;
  return 0;
}
void RefreshMRT(UINT32 t,ExportListStruct *e)
{
  if (e == NULL) return;
  if (t == e->almMRT) 
  {
    e->nAlmMRT++;
  }
  else 
  {
    e->almMRCT = e->almMRT;
    e->nAlmMRCT = e->nAlmMRT;
    e->almMRT = t;
    e->nAlmMRT = 1;
  }
}
void RefreshHSV(short s,ExportListStruct *e) 
{
  if (e == NULL) return;
  if (s == e->almHSV)
  {
    e->nAlmHSV++;
  }
  else if (s > e->almHSV)
  { 
    e->almHSV = (BYTE)s; 
    e->nAlmHSV = 1; 
  }
}
int SetAlarmByDeviceNumber(char *name,short devNr,UINT32 almCode,BYTE *almData,BYTE flags,double ts)
{
  short swpfmt,swpdim,cc = 0,adatasize = 0;
  ALARM *alm = NULL;
  ALARM *lsa = NULL;  /* least significant alarm */
  short lsc,sc = findSeverity(name,almCode),tmpsc = sc;
  ADS *ads;
  ExportListStruct *el;

  if (!AlarmServerInitialized) return -1;
  if (AlarmTestState) flags |= almTEST; /* declared test state */
  /* Instant alarms pickup the TERMINATE signal immediately */
  if (flags & almINSTANT) flags |= almTERMINATE;
  /* fill in ALARMStruct for the tag via link list: */
  if ((el=getExportListItem(name)) == NULL) ccerr(0);
  if (verifyDeviceNr(el,devNr)) ccerr(out_of_range);
  if ((ads=getAlarmTable(name,almCode)) != NULL && almData != NULL)
  {
    if ((swpfmt=LFMT(ads->alarmDataFormat)) == CF_BIT)
    {
      swpfmt = CF_BYTE;
      swpdim = MIN((ads->alarmDataArraySize+7)/8,ALARM_DATA_SIZE);
    }
    else 
    {
      swpdim = MIN(ads->alarmDataArraySize,ALARM_DATA_SIZE);
    }
    adatasize = swpdim*getFormatSize(swpfmt);
    prepOutgoingAlarmData(almData,swpdim,swpfmt);
  }
  for (alm=el->almLst[devNr]; alm != NULL; alm=alm->next)
  {
    if ((lsc=findSeverity(name,alm->alarmCode)) < tmpsc) { lsa = alm; tmpsc = lsc; }
    if (almCode == alm->alarmCode &&          /* it's in the list */
        !(alm->alarmStatus & almTERMINATE))   /* and not marked for termination */
    {
      if (alm->alarmStatus & almNEWALARM && 
          ts <= tval2ts(alm->timestamp,alm->timestampUSec) + NEW_ALARM_WINDOW) /* a double SetAlarm() ? */
      {
        alm->clrCount = 0;                     /* reset clear count      */
        return 0;
      }
      if (almData != NULL && memcmp(almData,alm->alarmData,adatasize))
      {                                        /* alarm data have changed */
        alm->alarmStatus &= ~((BYTE)almNEWALARM); /* remove NEWALARM */
        alm->alarmStatus |= almDATACHANGE;
        AssignAlmTimestamp(alm,ts);
        RefreshMRT(alm->timestamp,el);
        memcpy(alm->alarmData,almData,adatasize); /* copy the new data ! */
        if (ALMdebug)
        {
          dbglog("ALM: datachange set for %s <%d> (0x%x)",
            el->EqmName,(int)alm->alarmCode,alm->alarmStatus);
        }
      }
      else
      {                                        /* alarm has been reset */
        if (alm->clrCount > 1)
        {
          alm->alarmStatus |= almOSCILLATION;  /* appears to be oscillating */
          if (ALMdebug)
          {
            dbglog("ALM: oscillation set for %s <%d> (0x%x)",
              el->EqmName,(int)alm->alarmCode,alm->alarmStatus);
          }
          alm->alarmStatus &= ~((BYTE)almNEWALARM); /* remove NEWALARM */
          AssignAlmTimestamp(alm,ts);
          RefreshMRT(alm->timestamp,el);
        }
      }
      alm->clrCount = 0;                     /* reset clear count      */
      return 0;                              /* already in list, so return */
    }
  }
  flags |= almNEWALARM;                       /* alarm not in list */
  if (el->nalarms < MaxNumberAlarms)          /* still room to put the alarm in list */
  {
    if ((alm=(ALARM *)SystemCalloc(1,sizeof(ALARM))) == NULL) ccerr(out_of_server_memory);
    if (gRequireReadFromCAS) alm->removalBlocked = TRUE; /* must be cleared by CAS readout */
    AssignAlmTimestamp(alm,ts);
    alm->alarmCode = almCode;
    alm->alarmStatus |= flags;
    if (adatasize > 0) memcpy(alm->alarmData,almData,adatasize);
    alm->next = (ALARM *)el->almLst[devNr];
    el->almLst[devNr] = alm;
    el->nalarms++;
    RefreshMRT(alm->timestamp,el);
    RefreshHSV(sc,el);
    if (ALMdebug)
      dbglog("ALM: alarm set for %s <%d> (0x%x)",el->EqmName,(int)alm->alarmCode,alm->alarmStatus);
    return 0;
  }
  else if (lsa != NULL)  /* replace the least significant alarm */
  {
    alm = lsa;
    if (ALMdebug)
    {
      dbglog("ALM: replace alarm %s (%d) with %s (%d)",
	      el->EqmName,(int)alm->alarmCode,
	      el->EqmName,(int)almCode);
    }
    AssignAlmTimestamp(alm,ts);
    alm->alarmCode = almCode;
    alm->alarmStatus = flags;
    if (adatasize > 0)
    {
      memset(alm->alarmData,0,ALARM_DATA_SIZE);
      memcpy(alm->alarmData,almData,adatasize);
    }
    RefreshMRT(alm->timestamp,el);
    RefreshHSV(sc,el);
  }
  alm->starttime = alm->timestamp;
  alm->starttimeUSec = alm->timestampUSec;
  ccerr(max_alarms_exceeded);
err:
  switch (cc)
  {
    case 0: return 0;
    case out_of_server_memory:
    case max_alarms_exceeded:
      enforceAlarm(cc);
      break;
    default:
      if (alm) SystemFree((char *)alm);
      break;
  }
  return cc;
}
/* and in the background... */
ALARM *RemoveAlarmFromList(ExportListStruct *el, int n, ALARM *alm)
{
  ALARM *a0, *a1, *tmp = alm;

  if (el == NULL) return alm->next;
  if (n >= el->EqmNumDevices) return alm->next;
  for (a1=a0=el->almLst[n]; a1 != NULL; a0=a1,a1=a1->next)
  {
    if (a1 == alm) /* found it */
    {
      if (a1 == a0) el->almLst[n] = a1->next;
      else a0->next = a1->next;
      tmp = a1->next;
      if (a1->timestamp == el->almMRT) el->nAlmMRT--;
      if (a1->timestamp == el->almMRCT) el->nAlmMRCT--;
      if (el->almHSV == findSeverity(el->EqmName,a1->alarmCode))
      {
        if (--(el->nAlmHSV) <= 0)
        {
          el->nAlmHSV = 0; /* reset it */
          el->almHSV = 0;
        }
      }
      SystemFree((char *)a1); a1 = NULL;
      el->nalarms--;

      break;
    }
  }
  return tmp;    /* return the next link in the list */
}

void SurveyAlarms(struct timeval *tv)
{
  short sc;
  int n, findHSV;
  ALARM *alm;
  ExportListStruct *el;
  static struct timeval tv_interim;

  if (!AlarmServerInitialized) return;
  if (MSECS(*tv,tv_interim) < gAlarmSurveyInterval) return;
  if (ServerExitCondition) return;
  gettimeofday(&tv_interim,(struct timezone *)NULL);

  if (nAlmWatchList > 0) checkAlarmsInWatchTable();
  for (el=ExportList; el != NULL; el=el->next)
  {
    if (tv->tv_sec > (long)el->almMRT) /* adjust most recent complete timestamp */
    {
      el->almMRCT = el->almMRT;
      el->nAlmMRCT = el->nAlmMRT;
    }
    findHSV = (el->almHSV == 0) ? TRUE : FALSE;
    for (n=0;el->nalarms && n<el->EqmNumDevices; n++)
    {
      for (alm=el->almLst[n]; alm != NULL;)
      {
        if ((alm->alarmStatus & almTERMINATE)) /*  time to clear the alarm ? */
        {
          if (alm->removalBlocked) 
          {
            alm = alm->next;
            continue;
          }
          if (alm->timestamp + TerminationWindow < (UINT32)tv->tv_sec) /* okay to clear the alarm */
          {
            if (ALMdebug)
              dbglog("ALM: clear %s <%d> (0x%x)",el->EqmName,(int)alm->alarmCode,alm->alarmStatus);
            alm = RemoveAlarmFromList(el,n,alm);
            continue;
          }
          else if ((alm->alarmStatus & almINSTANT) &&
                    alm->timestamp + AlarmInstantWindow < (UINT32)tv->tv_sec)
          {
            if (ALMdebug)
              dbglog("ALM: clear Instant alarm %s <%d> (0x%x)",el->EqmName,(int)alm->alarmCode,alm->alarmStatus);
            alm = RemoveAlarmFromList(el,n,alm);
            continue;
          }
        }
        else                                   /* check alarm status */
        {
          if ((alm->clrCount > 3) &&
               alm->timestamp + AlarmWindow < (UINT32)tv->tv_sec)
          {
            if (ALMdebug)
              dbglog("ALM: termination for %s <%d> (0x%x)",el->EqmName,(int)alm->alarmCode,alm->alarmStatus);
            alm->alarmStatus &= ~((BYTE)almNEWALARM); /* remove NEWALARM */
            alm->alarmStatus |= almTERMINATE;
            alm->timestamp = tv->tv_sec;
            RefreshMRT(alm->timestamp,el);
          }
          else if (alm->timestamp + AlarmHeartbeat < (UINT32)tv->tv_sec) /*  set the heartbeat */
          {
            if (ALMdebug)
              dbglog("ALM: heartbeat set for %s <%d> (0x%x)",el->EqmName,(int)alm->alarmCode,alm->alarmStatus);
            alm->alarmStatus &= ~((BYTE)almNEWALARM); /* remove NEWALARM */
            alm->alarmStatus |= almHEARTBEAT;
            alm->timestamp = tv->tv_sec;
            RefreshMRT(alm->timestamp,el);
          }
        }
        if (findHSV)
        {
          if ((sc=findSeverity(el->EqmName,alm->alarmCode)) > el->almHSV)
          {
            el->almHSV = sc; el->nAlmHSV = 1;
          }
          else if (sc > 0 && sc == el->almHSV) el->nAlmHSV++;
        }
        alm = alm->next;
      }
    }
  }
  return;
}
/* caller subscribes to NALARMS which calls (REFRESH mode) */
int numAlarmsEx(char *eqmName,UINT32 *almTimeStamp,short *almSeverity,short *NrAtAlmTimeStamp,short *NrAtAlmSeverity)
{
  ExportListStruct *el;

  *almTimeStamp = 0; *almSeverity = 0;

  if ((el=getExportListItem(eqmName)) == NULL) return 0;
  *almTimeStamp = el->almMRCT;
  *almSeverity  = el->almHSV;
  *NrAtAlmTimeStamp = el->nAlmMRCT;
  *NrAtAlmSeverity = el->nAlmHSV;
  return el->nalarms;
}
int numAlarms(char *eqmName,UINT32 *almTimeStamp,short *almSeverity)
{
  return numAlarmsEx(eqmName,almTimeStamp,almSeverity,NULL,NULL);
}
int numAlarmDefs(char *eqm)
{
  ExportListStruct *el;

  if ((el=getExportListItem(eqm)) == NULL) return -non_existent_elem;
  return el->adsTableSize;
}
int getAlarmDefs(char *eqm,ADSList **adslst)
{
  ExportListStruct *el;

  if ((el=getExportListItem(eqm)) == NULL) return 0;
  *adslst = el->adsTable;
  return el->adsTableSize;
}
short findSeverity(char *eqm, UINT32 cc)
{
  ADSList *lst;
  ExportListStruct *start,*stop,*el;

  if (eqm == NULL)
  {
    start = ExportList; stop = NULL;
  }
  else
  {
    if ((start=getExportListItem(eqm)) == NULL) return 0;
    stop = start->next;
  }
  for (el=start; el != stop; el=el->next)
  {
    if ((lst=el->adsTable) != NULL) 
    {
      while (lst != NULL && lst->ads.alarmCode != (UINT32)cc) lst = lst->next;
      if (lst != NULL) return lst->ads.alarmSeverity;
    }
  }
  switch (cc)
  {
    case invalid_data: return almWatchTableADS[idx_invalid_data].alarmSeverity;
    case value_too_high: return almWatchTableADS[idx_too_high].alarmSeverity;
    case value_too_low: return almWatchTableADS[idx_too_low].alarmSeverity;
    case warn_too_high: return almWatchTableADS[idx_warn_too_high].alarmSeverity;
    case warn_too_low: return almWatchTableADS[idx_warn_too_low].alarmSeverity;
  }
  return 0;
}
ADS *getAlarmTable(char *eqm,UINT32 code)
{
  ADSList *lst;
  ExportListStruct *el;

  if ((el=getExportListItem(eqm)) != NULL)
  {
    for (lst=el->adsTable; lst != NULL; lst=lst->next) if ((UINT32)code == lst->ads.alarmCode) return &lst->ads;
    switch (code)
    {
      case invalid_data: return &almWatchTableADS[idx_invalid_data];
      case value_too_high: return &almWatchTableADS[idx_too_high];
      case value_too_low: return &almWatchTableADS[idx_too_low];
      case warn_too_high: return &almWatchTableADS[idx_warn_too_high];
      case warn_too_low: return &almWatchTableADS[idx_warn_too_low];
    }
  }
  return NULL;
}
TINE_EXPORT int getAlarmsEx(char *eqmName,AMS *ams,short num,short minSeverity,time_t starttime,time_t stoptime,short startnum)
{
  int sc,dv,i=0,n=0;
  ALARM *alm;
  ADS *ads;
  ExportListStruct *el;
  char defaultDeviceName[DEVICE_NAME_SIZE+4];

  if ((el=getExportListItem(eqmName)) == NULL) return 0;
  for (dv=0; el->nalarms && dv<el->EqmNumDevices; dv++)
  {
    for (alm=el->almLst[dv]; alm != NULL; alm=alm->next)
    {
      if (alm->removalBlocked && isCASInUserList(eqmName)) alm->removalBlocked = FALSE;
      if ((time_t)alm->timestamp >= starttime && (time_t)alm->timestamp <= stoptime &&
          (sc=findSeverity(eqmName,alm->alarmCode)) >= minSeverity )
      {
        if (i++ < startnum) continue;
        /* fill in ams[n] here ... */
        strncpy(ams[n].server,el->EqmExportName,EXPORT_NAME_SIZE);
        sprintf(defaultDeviceName,"DEVICE %d",dv);
        if (el->EqmDeviceList) strncpy(ams[n].device,el->EqmDeviceList[dv].name,DEVICE_NAME_SIZE);
        if (strlen(ams[n].device) == 0) strncpy(ams[n].device,defaultDeviceName,DEVICE_NAME_SIZE);
        if ((ads=getAlarmTable(eqmName,alm->alarmCode)) != NULL)
        {
          strncpy(ams[n].alarmTag,ads->alarmTag,ALARM_TAG_SIZE);
          ams[n].alarmMask = ads->alarmMask;
          ams[n].alarmDataFormat = ads->alarmDataFormat;
          ams[n].alarmDataArraySize = ads->alarmDataArraySize;
          ams[n].alarmSystem = ads->alarmSystem;
        }
        ams[n].timestamp = alm->timestamp;
        ams[n].timestampUSec = alm->timestampUSec;
        ams[n].starttime = alm->starttime;
        ams[n].starttimeUSec = alm->starttimeUSec;
        ams[n].alarmCode = alm->alarmCode;
        memcpy(ams[n].alarmData,alm->alarmData,ALARM_DATA_SIZE);
        ams[n].severity = sc;
        ams[n].descriptor = alm->alarmStatus;
        if (ams[n].alarmCode < (UINT32)numErr && strlen(ams[n].alarmTag) == 0)
          strncpy(ams[n].alarmTag,erlst[ams[n].alarmCode],ALARM_TAG_SIZE);
        if (++n >= num) return n;
      }
    }
  }
  return n;
}
TINE_EXPORT int getAlarms(char *eqmName,AMS *ams,short num,short minSeverity)
{
  return getAlarmsEx(eqmName,ams,num,minSeverity,0,0x7fffffffL,0);
}
TINE_EXPORT int GetAlarm(char *eqmName,char *devname,AMS *ams) /* get highest alarm for given device */
{
  int sc,scmax=-1,dv;
  ALARM *alm;
  ADS *ads;
  ExportListStruct *el;

  if ((el=getExportListItem(eqmName)) == NULL) return non_existent;
  if ((dv=GetDeviceNumber(eqmName,devname)) < 0) return illegal_equipment_number;
  for (alm=el->almLst[dv]; alm != NULL; alm=alm->next)
  {
    if ((sc=findSeverity(eqmName,alm->alarmCode)) > scmax)
    {
      scmax = sc;
      /* fill in ams[n] here ... */
      strncpy(ams->server,el->EqmExportName,EXPORT_NAME_SIZE);
      strncpy(ams->device,devname,DEVICE_NAME_SIZE);
      if ((ads=getAlarmTable(eqmName,alm->alarmCode)) != NULL)
      {
        strncpy(ams->alarmTag,ads->alarmTag,ALARM_TAG_SIZE);
        ams->alarmMask = ads->alarmMask;
        ams->alarmDataFormat = ads->alarmDataFormat;
        ams->alarmDataArraySize = ads->alarmDataArraySize;
        ams->alarmSystem = ads->alarmSystem;
      }
      ams->timestamp = alm->timestamp;
      ams->timestampUSec = alm->timestampUSec;
      ams->starttime = alm->starttime;
      ams->starttimeUSec = alm->starttimeUSec;
      ams->alarmCode = alm->alarmCode;
      memcpy(ams->alarmData,alm->alarmData,ALARM_DATA_SIZE);
      ams->severity = sc;
      ams->descriptor = alm->alarmStatus;
    }
  }
  return 0;
}

int prepIncomingAMRB(AMRB *amrb,BYTE *amrbstream) 
{ /* structure is an array of long ints */
  memcpy(amrb,amrbstream,AMRB_SIZE);
# ifndef INTEL
  SwapBytes((char *)amrb,4,CF_LONG,RTOL);
# endif
  return 0;
}
int prepOutgoingAMS(BYTE *amsstream,AMS *amslst,char *stag,int length)  
{
  AMS5 ams5;
  int i;
  if (!strcmp(stag,"AMSr4"))
  { /* modern alarm definition structure */
#    ifdef INTEL
      if ((void *)amsstream != (void *)amslst) memcpy(amsstream,amslst,length*AMS_SIZE); 
#    else
     _StructSwapEx(AMSStructStruct,amsstream,(unsigned char *)amslst,LTOR,length,sizeof(AMS));
#    endif
  }
  else
  { /* old style */
    for (i=0; i<length; i++)
    {
      memset(&ams5,0,sizeof(AMS5));
      strncpy(ams5.server,amslst[i].server,EXPORT_NAME_SHORTSIZE);
      strncpy(ams5.device,amslst[i].device,DEVICE_NAME_SHORTSIZE);
      strncpy(ams5.alarmTag,amslst[i].alarmTag,ALARM_TAG_SHORTSIZE);
      ams5.alarmCode = amslst[i].alarmCode;
      ams5.timestamp = amslst[i].timestamp;
      ams5.alarmMask = amslst[i].alarmMask;
      memcpy(ams5.alarmData,amslst[i].alarmData,ALARM_DATA_SHORTSIZE);
      ams5.alarmDataFormat = amslst[i].alarmDataFormat;
      ams5.alarmDataArraySize = amslst[i].alarmDataArraySize;
      ams5.severity = amslst[i].severity;
      ams5.descriptor = amslst[i].descriptor;
      ams5.alarmSystem = amslst[i].alarmSystem;
#     ifdef INTEL
      memcpy(&amsstream[i*AMS5_SIZE],&ams5,AMS5_SIZE);
#     else
      _StructSwapEx(AMS5StructStruct,&amsstream[i*AMS5_SIZE],(unsigned char *)&ams5,LTOR,1,sizeof(AMS5));
#     endif
    }
  }
  return 0;
}
int prepOutgoingAlarmData(BYTE *datastream,int size,short fmt)
{
# ifdef INTEL
  datastream = datastream;
  size = size;
  fmt = fmt;
# else
  SwapBytes((char *)datastream,size,fmt,LTOR);
# endif
  return 0;
}
int prepIncomingAWS(BYTE *awsstream,AWS *aws,char *stag,int length)
{
  if (strcmp(stag,"AWSr4") != 0) return invalid_structure_tag;
# ifdef INTEL
  memcpy(aws,awsstream,length*AWS_SIZE);
# else
  _StructSwapEx(AWSStructStruct,awsstream,(unsigned char *)aws,RTOL,length,sizeof(AWS));
# endif
  return 0;
}
int prepOutgoingAWS(BYTE *awsstream,AWSList *awslst,char *stag,int length)
{
  int i;
  AWSList *aw;
  if (strcmp(stag,"AWSr4") != 0) return invalid_structure_tag;
  for (i=0, aw=awslst; i<length  && aw != NULL; i++, aw=aw->nxt)
  {
#   ifdef INTEL
    memcpy(&awsstream[i*AWS_SIZE],&aw->aws,AWS_SIZE);
#   else
    _StructSwapEx(AWSStructStruct,&awsstream[i*AWS_SIZE],(unsigned char *)&aw->aws,LTOR,1,sizeof(AWS));
#   endif
  }
  return 0;
}

int prepOutgoingADS(BYTE *adsstream,ADSList *adslst,char *stag,int length)
{
  int i;
  ADS5 ads5;
  ADSList *lst;

  if (!strcmp(stag,"ADSr4"))
  { /* modern alarm definition structure */
    for (i=0, lst=adslst; i<length && lst != NULL; i++, lst=lst->next)
    {
#     ifdef INTEL
      memcpy(&adsstream[i*ADS_SIZE],&lst->ads,ADS_SIZE);
#     else
      _StructSwapEx(ADSStructStruct,&adsstream[i*ADS_SIZE],(unsigned char *)&lst->ads,LTOR,1,sizeof(ADS));
#     endif
    }
  }
  else
  { /* old style */
    for (i=0, lst=adslst; i<length && lst != NULL; i++, lst=lst->next)
    {
      memset(&ads5,0,sizeof(ADS5));
      ads5.alarmCode = lst->ads.alarmCode;
      ads5.alarmDataArraySize = lst->ads.alarmDataArraySize;
      ads5.alarmDataFormat = lst->ads.alarmDataFormat;
      ads5.alarmMask = lst->ads.alarmMask;
      ads5.alarmSystem = lst->ads.alarmSystem;
      strncpy(ads5.alarmTag,lst->ads.alarmTag,ALARM_TAG_SHORTSIZE);
      strncpy(ads5.alarmText,lst->ads.alarmText,ALARM_TEXT_SHORTSIZE);
      strncpy(ads5.deviceText,lst->ads.deviceText,ALARM_TEXT_SHORTSIZE);
      ads5.severity = lst->ads.alarmSeverity;
      strncpy(ads5.url,lst->ads.url,ALARM_TEXT_SHORTSIZE);
#     ifdef INTEL
      memcpy(&adsstream[i*ADS5_SIZE],&ads5,ADS5_SIZE);
#     else
      _StructSwapEx(ADS5StructStruct,&adsstream[i*ADS5_SIZE],(unsigned char *)&ads5,LTOR,1,sizeof(ADS5));
#     endif
    }
  }
  return 0;
}

#endif /* NO_DOXYGEN */

/* Local Alarm Server API Calls follow: */
TINE_EXPORT int AlarmWindow = DEFAULT_ALARM_WINDOW;
TINE_EXPORT int AlarmHeartbeat = DEFAULT_ALARM_HEARTBEAT;
TINE_EXPORT int AlarmInstantWindow = DEFAULT_INSTANT_WINDOW;
TINE_EXPORT int MaxNumberAlarms = MAXALARMS;
TINE_EXPORT int TerminationWindow =  DEFAULT_TERMINATION_WINDOW;
/* provide API registration as alternative to -alarms.csv file : */
#ifndef SMALL_TINE_LIB /* for DOS, etc. ... */
TINE_EXPORT int AppendAlarmInfoTable(char *eqm,ADS *ads)
{
  ExportListStruct *el;
  ADSList *lst=NULL;
  int cc=0;
  
  if (alarmInfoTableFile) ccerr(not_allowed);
  if (eqm == NULL || ads == NULL) ccerr(argument_list_error);
  if ((el=getExportListItem(eqm)) == NULL) ccerr(non_existent_elem);
  if ((lst=(ADSList *)SystemCalloc(1,sizeof(ADSList))) == NULL) ccerr(out_of_server_memory);
  memcpy(&lst->ads,ads,sizeof(ADS));
  lst->next = el->adsTable;
  el->adsTable = lst;
  el->adsTableSize++;
  if (lst->ads.alarmSystem) el->alarmSystem = 1;
err:
  return cc;
}
TINE_EXPORT int AppendAlarmWatchTable(char *eqm,char *prp,char *dev,int siz,int fmt,int atyp,int sev,int sys,ALM_THRESHOLDS *thr)
{
  int cc = 0;
  AWSList *aw;

  if (eqm == NULL || prp == NULL || dev == NULL || thr == NULL) ccerr(argument_list_error);
  if ((aw=(AWSList *)SystemCalloc(1,sizeof(AWSList))) == NULL) ccerr(out_of_server_memory);
  strncpy(aw->aws.eqm,eqm,EQM_NAME_SHORTSIZE);
  strncpy(aw->aws.prp,prp,PROPERTY_NAME_SIZE);
  strncpy(aw->aws.dev,dev,DEVICE_NAME_SIZE);
  aw->aws.siz = siz;
  aw->aws.fmt = fmt;
  aw->aws.atyp = atyp;
  aw->aws.sev = sev;
  aw->aws.asys = sys;
  aw->aws.cntThreshold = thr->cnt;
  if (thr->mask != 0)
  {
    aw->aws.mask = thr->mask;
    aw->aws.normal = thr->normal;
    aw->aws.chkNormal = TRUE;
  }
  aw->aws.hi = thr->hi;
  aw->aws.hiwarn = thr->hiwarn;
  aw->aws.lo = thr->lo;
  aw->aws.lowarn = thr->lowarn;
  aw->aws.hisev = thr->hisev > 0 ? thr->hisev : aw->aws.sev;
  aw->aws.hiwarnsev = thr->hiwarnsev > 0 ? thr->hiwarnsev : aw->aws.hisev - 2;
  aw->aws.losev = thr->losev > 0 ? thr->losev : aw->aws.sev;
  aw->aws.lowarnsev = thr->lowarnsev > 0 ? thr->lowarnsev : aw->aws.losev - 2;
  aw->nxt = almWatchList;
  almWatchList = aw;
  nAlmWatchList++;
  feclog("LAS: append %s %s to alarm watch table",dev,prp);
err:
  return cc;
}
#endif /* SMALL_TINE_LIB */
TINE_EXPORT void ClearAlarm(char *name,short devNr)
{
  ALARM *alm;
  ExportListStruct *el;
  int start,stop,i;

  if (!AlarmServerInitialized) return;

  if ((el=getExportListItem(name)) == NULL) return;
  if (devNr == -1)                        /* clear them all */
  {
    start = 0; stop = el->EqmNumDevices;
  }
  else
  {
    if (verifyDeviceNr(el,devNr)) return;
    start = devNr; stop = devNr + 1;
  }
  for (i=start; i<stop; i++)
  {
    for (alm=el->almLst[i]; alm != NULL; alm=alm->next)
    {
      if (alm->alarmStatus & almINSTANT) alm->clrCount = 8; /* it'll be cleared immediately */
      else alm->clrCount = MIN(alm->clrCount+1,8); /* increment the clear count */
    }
  }
  return;                                 /* wasn't set */
}
TINE_EXPORT void ClearDeviceAlarm(char *name,char *devname)
{
  short devNr = -1;
  if (strcmp(devname,"ALL")) devNr = findDeviceNr(name,devname);
  ClearAlarm(name,devNr);
}
TINE_EXPORT int GetAlarmHeartbeat(void) { return AlarmHeartbeat; }
TINE_EXPORT int GetAlarmTableMaximumSize(void) { return MaxNumberAlarms; }
TINE_EXPORT int GetAlarmTerminationWindow(void) { return TerminationWindow; }
TINE_EXPORT int GetNewAlarmWindow(void) { return AlarmWindow; }
TINE_EXPORT void RemoveAlarm(char *name,short devNr,UINT32 code)  /* clear it immediately */
{
  ALARM *alm;
  ExportListStruct *el;

  if (!AlarmServerInitialized) return;
  if ((el=getExportListItem(name)) == NULL) return;
  if (verifyDeviceNr(el,devNr)) return;

  for (alm=el->almLst[devNr]; alm != NULL; alm=alm->next)
  {
    if (code == alm->alarmCode)
    {
      alm->clrCount = 8;
    }
  }
  return;                                 /* wasn't set */
}
TINE_EXPORT void RemoveDeviceAlarm(char *name,char *devname,UINT32 code)  /* clear it immediately */
{
  short devNr = findDeviceNr(name,devname);
  RemoveAlarm(name,devNr,code);
}
TINE_EXPORT int SetAlarm(char *name,short devNr,UINT32 almCode,BYTE *almData)
{
  return SetAlarmEx(name,devNr,almCode,almData,almNEWALARM);
}
TINE_EXPORT int SetAlarmEx(char *name,short devNr,UINT32 almCode,BYTE *almData,BYTE flags)
{
  double ts = makeDataTimeStamp();
  return SetAlarmByDeviceNumber(name,devNr,almCode,almData,flags,ts);
}
TINE_EXPORT void SetAlarmHeartbeat(int timeInSeconds) 
{ 
  if (timeInSeconds < 5) return;
  AlarmHeartbeat = timeInSeconds; 
}
TINE_EXPORT void SetAlarmTableMaximumSize(int tableSize) 
{ 
  if (tableSize < 10) return;
  MaxNumberAlarms = tableSize; 
}
TINE_EXPORT void SetAlarmTerminationWindow(int timeInSeconds) 
{ 
  if (timeInSeconds < 5) return;
  TerminationWindow = timeInSeconds; 
}
TINE_EXPORT void SetNewAlarmWindow(int timeInSeconds) 
{ 
  if (timeInSeconds < 1) return;
  AlarmWindow = timeInSeconds; 
}
TINE_EXPORT int SetDeviceAlarm(char *name,char *devname,UINT32 almCode,BYTE *almData,BYTE flags)
{
  short devNr = findDeviceNr(name,devname);
  return SetAlarmEx(name,devNr,almCode,almData,flags);
}
